/**************************************************************************************************
	$Header: /pub/cvsroot/mydns/src/mydns/udp.c,v 1.20 2003/07/25 09:26:45 bboy Exp $

	Copyright (C) 2002, 2003  Don Moore <bboy@bboy.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at Your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************************************/

#include "named.h"


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	READ_UDP_QUERY
	Returns 0 on success (a task was added), -1 on failure.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
read_udp_query(int fd)
{
	struct sockaddr_in addr;
	char		in[DNS_MAXPACKETLEN_UDP], *src, *qdtop;
	int		addrlen, len;
	char		qname[DNS_MAXNAMELEN+1];
	TASK		*t;
	uint16_t	ancount, nscount, arcount;

	/* Read message */
	addrlen = sizeof(addr);
	memset(&in, 0, sizeof(in));
	if ((len = recvfrom(fd, &in, sizeof(in), 0, (struct sockaddr *)&addr, &addrlen)) == -1)
		return (-1);

	if (len < DNS_HEADERSIZE)
		return Warnx("%s: %s", inet_ntoa(addr.sin_addr), _("UDP message too short"));
	if (len > DNS_MAXPACKETLEN_UDP)
		return Warnx("%s: %s", inet_ntoa(addr.sin_addr), _("UDP message too long"));

	/* Start new task */
	t = task_init(NEED_ANSWER, fd, SOCK_DGRAM, &addr);

#if DEBUG_ENABLED
	Debug("[%lu] %s: %d UDP octets in", t->internal_id, inet_ntoa(t->addr.sin_addr), len);
#endif

	/* Parse query header data */
	src = in;
	DNS_GET16(t->id, src);
	memcpy(&t->hdr, src, SIZE16); src += SIZE16;
	DNS_GET16(t->qdcount, src);
	DNS_GET16(ancount, src);
	DNS_GET16(nscount, src);
	DNS_GET16(arcount, src);

#if DEBUG_ENABLED
	Debug("%s: id=%u qr=%u opcode=%s aa=%u tc=%u rd=%u ra=%u z=%u rcode=%u", desctask(t),
			t->id, t->hdr.qr, mydns_opcode_str(t->hdr.opcode),
			t->hdr.aa, t->hdr.tc, t->hdr.rd, t->hdr.ra, t->hdr.z, t->hdr.rcode);
	Debug("%s: qd=%u an=%u ns=%u ar=%u", desctask(t),
			t->qdcount, ancount, nscount, arcount);
#endif

	/* Check validity of header fields */
	if (!ntohs(t->qdcount))
		return formerr(t, DNS_R_FORMERR, ERR_NO_QUESTION);

	if (t->hdr.tc)
		return formerr(t, DNS_R_FORMERR, ERR_QUESTION_TRUNCATED);

	if (ntohs(t->hdr.opcode) != DNS_OPCODE_QUERY)
		return formerr(t, DNS_R_NOTIMP, ERR_UNSUPPORTED_OPCODE);

	/* Fill in question data */
	t->qdlen = len - DNS_HEADERSIZE;
	if (t->qdlen <= 0)
		return formerr(t, DNS_R_FORMERR, ERR_MALFORMED_REQUEST);

	if (!(t->qd = (char *)malloc(t->qdlen)))
		Err("malloc");
	memcpy(t->qd, src, t->qdlen);
	qdtop = src;

	/* Get query name */
	if (!(src = name_unencode(t->qd, t->qdlen, src, qname, sizeof(qname))))
		return formerr(t, DNS_R_FORMERR, qname[0]);
	strncpy(t->qname, qname, sizeof(t->qname)-1);

	/* Get query type */
	if (src + SIZE16 > in + len)
		return formerr(t, DNS_R_FORMERR, ERR_MALFORMED_REQUEST);
	DNS_GET16(t->qtype, src);

	/* Get query class */
	if (src + SIZE16 > in + len)
		return formerr(t, DNS_R_FORMERR, ERR_MALFORMED_REQUEST);
	DNS_GET16(t->qclass, src);

	t->qdlen = src - qdtop;

	/* Check class (only IN or ANY are allowed) */
	if ((t->qclass != DNS_C_IN) && (t->qclass != DNS_C_ANY))
	{
		dnserror(t, DNS_R_NOTIMP, ERR_NO_CLASS);
		build_reply(t, 1);
		return (-1);
	}

	/* Enqueue new item */
	if (enqueue(Tasks, t) < 0)
	{
		task_free(t);
		return (-1);
	}

	return (0);
}
/*--- read_udp_query() --------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	WRITE_UDP_REPLY
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
write_udp_reply(TASK *t)
{
	int rv;

	if ((rv = sendto(t->fd, t->reply, t->replylen, MSG_DONTWAIT,
					(struct sockaddr *)&t->addr, sizeof(t->addr))) < 0)
	{
		Warn("%s: %s", desctask(t), _("error writing reply"));
		dequeue(Tasks, t);
		return (-1);
	}

	/* Task complete; dequeue */
#if DEBUG_ENABLED
	Debug("%s: WRITE %u UDP octets (id %u)", desctask(t), t->replylen, t->id);
#endif
	dequeue(Tasks, t);

	return (0);
}
/*--- write_udp_reply() -------------------------------------------------------------------------*/

/* vi:set ts=3: */
/* NEED_PO */
